/*
 * HDLC driver support.
 *
 * Copyright (C) 2002 Network Systems Group, Igor Mosunov <igor@nsg.ru>
 *
 * This file is part of GNU Zebra.
 *
 * GNU Zebra is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * GNU Zebra is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Zebra; see the file COPYING.  If not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#include <zebra.h>

#include "if.h"
#include "vty.h"
#include "sockunion.h"
#include "prefix.h"
#include "command.h"
#include "memory.h"
#include "ioctl.h"
#include "connected.h"
#include "log.h"
#include "zclient.h"

#include "zebra/interface.h"
#include "zebra/rtadv.h"
#include "zebra/rib.h"
#include "zebra/zserv.h"
#include "zebra/redistribute.h"
#include "zebra/debug.h"
#include "zebra/nsg_hdlc.h"

/* Master list of nsg_ports. */
struct list *nsg_port_list;


nsg_value_t nsg_port_encap_value[] = {
  {CISCO_HDLC_STR, IF_PROTO_CISCO},
  {FRAME_RELAY_STR, IF_PROTO_FR},
  {NULL}
};

nsg_value_t nsg_port_fr_lmi_value[] = {
  {NONE_STR, LMI_NONE},
  {ANSI_STR, LMI_ANSI},
  {CCITT_STR, LMI_CCITT},
  {NULL}
};

nsg_value_t nsg_port_fr_intf_type_value[] = {
  {DTE_STR, 0},
  {DCE_STR, 1},
  {NULL}
};

nsg_value_t nsg_port_clock_value[] = {
  {EXTERNAL_STR, CLOCK_EXT},
  {INTERNAL_STR, CLOCK_INT},
  {MASTER_STR, MODEM_MASTER},
  {SLAVE_STR, MODEM_SLAVE},
  {NULL}
};

nsg_value_t nsg_port_interface_value[] = {
  {V24_STR, INTYPE_V24},
  {V35_STR, INTYPE_V35},
  {SDSL_STR, INTYPE_SDSL},
  {NULL}
};

nsg_port_config_t nsg_port_cisco_conf[] = {
  {KEEPALIVE_STR, DTYPE_UINT, offsetof(nsgdev_conf_t, param.cisco.interval), "10"},
  {TIMEOUT_STR,   DTYPE_UINT, offsetof(nsgdev_conf_t, param.cisco.timeout),  "25"},
  { NULL}
};

nsg_port_config_t nsg_port_fr_conf[] = {
  {LMI_STR,       DTYPE_USHORT, offsetof(nsgdev_conf_t, param.fr.lmi), NONE_STR, 
    nsg_port_fr_lmi_value},
  {INTF_TYPE_STR, DTYPE_USHORT, offsetof(nsgdev_conf_t, param.fr.dce), DTE_STR, 
    nsg_port_fr_intf_type_value},
  {T391_STR,      DTYPE_UINT, 	offsetof(nsgdev_conf_t, param.fr.t391), "10"},
  {T392_STR,      DTYPE_UINT,   offsetof(nsgdev_conf_t, param.fr.t392), "15"},
  {N391_STR,      DTYPE_UINT,	offsetof(nsgdev_conf_t, param.fr.n391), "6"},
  {N392_STR,      DTYPE_UINT,   offsetof(nsgdev_conf_t, param.fr.n392), "3"},
  {N393_STR,      DTYPE_UINT,   offsetof(nsgdev_conf_t, param.fr.n393), "4"},
  {NULL}
};

nsg_port_config_t nsg_port_conf[] = {
  {PHY_INTERFACE_STR, DTYPE_UINT, offsetof(nsgdev_conf_t, sync_serial.interface_type),
   V24_STR, nsg_port_interface_value},
  {CLOCK_STR, DTYPE_UINT, offsetof(nsgdev_conf_t, sync_serial.clock_type),
   EXTERNAL_STR, nsg_port_clock_value},
  {BAUDRATE_STR, DTYPE_UINT, offsetof(nsgdev_conf_t, sync_serial.clock_rate),
   "64000"},
  {ENCAPSULATION_STR, DTYPE_INT, offsetof(nsgdev_conf_t, encap), CISCO_HDLC_STR, 
   nsg_port_encap_value, 0, set_encap_param_hook},
  {CISCO_HDLC_STR, DTYPE_CONFIG,  0, 0, nsg_port_cisco_conf, 
   ENCAPSULATION_STR" "CISCO_HDLC_STR},
  {FRAME_RELAY_STR, DTYPE_CONFIG, 0, 0, nsg_port_fr_conf,
   ENCAPSULATION_STR" "FRAME_RELAY_STR},
  {NULL}
};

nsg_port_config_t nsg_port_entry_conf = 
  {"port", DTYPE_CONFIG, 0, 0, nsg_port_conf};


int get_key(char *name, char **key, char **nextname) {
  int leng;
  char *s;

  for (s = name; *s == ' '; s++);
  *key = s;
  for (; *s != ' ' && *s != '\0'; s++);
  leng = s - *key;
  for (; *s == ' '; s++);
  *nextname = s;
  return leng;
}


ulong get_value_by_name(char *name, int leng, nsg_value_t *value_table) {
  nsg_value_t *p;

  for (p = value_table; p->name; p++) {
    if (strncmp(name, p->name, leng) == 0) {
      errno = 0;
      return p->val;
    }
  }
  errno = ERANGE;
  return ULONG_MAX;
}

char *get_full_name(char *short_name, int leng, nsg_value_t *value_table) {
  nsg_value_t *p;

  for (p = value_table; p->name; p++) {
    if (strncmp(short_name, p->name, leng) == 0) {
      return p->name;
    }
  }
  return NULL;
}

int get_name_by_value(char *name, ulong val, nsg_value_t *value_table) {
  nsg_value_t *p;

  for (p = value_table; p->name; p++) {
    if (val == p->val) {
      strcpy(name, p->name);
      return 1;
    }
  }
  return 0;
}

int nsg_port_test_param(char *cond, nsg_port_config_t *confel, struct nsg_port_t *portp) {
  nsgdev_conf_t *conf = &portp->conf;
  ulong tmp = 0;
  char *nextcond;
  char *key;
  int  keyleng, ret;

  if (!cond || !(keyleng  = get_key(cond, &key, &nextcond))) {
    key = confel->default_value;
    keyleng = strlen(key);
    nextcond = 0;
  }

  if (!keyleng)
    return 1;

  if (confel->value) {
    tmp = get_value_by_name(key, keyleng, confel->value);
  }
  else {
    tmp = strtoul(key, NULL, 0);
  }
  if (errno == ERANGE)
    return 0;
  
  switch (confel->dtype) {
    
  case DTYPE_INT:
    ret = (*(int*)((char*)conf + confel->offset) == tmp);
    break;
    
  case DTYPE_UINT:
    ret = (*(uint*)((char*)conf + confel->offset) == tmp);
    break;
    
  case DTYPE_LONG:
    ret = (*(long*)((char*)conf + confel->offset) == tmp);
    break;
    
  case DTYPE_ULONG:
    ret = (*(ulong*)((char*)conf + confel->offset) == tmp);
    break;
    
  case DTYPE_SHORT:
    ret = (*(short*)((char*)conf + confel->offset) == tmp);
    break;
    
  case DTYPE_USHORT:
    ret = (*(ushort*)((char*)conf + confel->offset) == tmp);
    break;
    
  case DTYPE_CHAR:
    ret = (*(char*)((char*)conf + confel->offset) == tmp);
    break;
    
  case DTYPE_UCHAR:
    ret = (*(unsigned char*)((char*)conf + confel->offset) == tmp);
    break;
    
  case DTYPE_PTR:
    ret = (*(void**)((char*)conf + confel->offset) == (void*)tmp);
    break;
  }

  if (nextcond && nextcond[0])
    return 0;
  else
    return ret;
}

int nsg_port_test_cond(char *cond, nsg_port_config_t *confp, struct nsg_port_t *portp) {
  nsg_port_config_t *p;
  char *nextcond;
  char *key;
  int  keyleng;

  if (!cond) 
    return 1;

  keyleng  = get_key(cond, &key, &nextcond);

  if (!keyleng)
    return 1;

  for (p = confp; p->keyword; p++) {
    if (strncmp(key, p->keyword, keyleng) == 0) {
      if (p->dtype == DTYPE_CONFIG) {
	return nextcond[0] ? nsg_port_test_cond(nextcond, p->value, portp) : 1;
      }
      else 
	return nsg_port_test_param(nextcond, p, portp);
    }
  }
  return 0;

}

nsg_port_config_t *nsg_port_get_config(char *name, nsg_port_config_t *confp, 
				       struct nsg_port_t *portp) {
  nsg_port_config_t *p;
  char *nextname;
  char *key;
  int  keyleng  = get_key(name, &key, &nextname);

  if (!keyleng)
    return 0;

  for (p = confp; p->keyword; p++) {
    if (strncmp(key, p->keyword, keyleng) == 0) {
      if (!nsg_port_test_cond(p->cond, nsg_port_conf, portp))
	return 0;

      if (p->dtype == DTYPE_CONFIG) {
	return nextname[0] ? nsg_port_get_config(nextname, p->value, portp) : p;
      }
      else 
	return nextname[0] ? 0 : p;
    }
  }
  return 0;
}

int nsg_port_set_param(int argc, char **argv, 
		       nsg_port_config_t *confel, struct nsg_port_t *portp) {
  nsgdev_conf_t *conf = &portp->conf;
  ulong old,tmp = 0;
  char *name = argc ? argv[0] : confel->default_value;

  if (name == NULL || confel == NULL || 
      !nsg_port_test_cond(confel->cond, nsg_port_conf, portp))
    return 0;

  switch (confel->dtype) {
  case DTYPE_CONFIG:
    if (argc) {
      return 
	nsg_port_set_param(argc - 1, argv + 1, 
			   nsg_port_get_config(name,
					       (nsg_port_config_t *)confel->value,
					       portp),
					       portp);
    }
    else
      return 0;
    break;

  default:
    if (confel->value) {
      tmp = get_value_by_name(name, strlen(name), confel->value);
    }
    else {
      tmp = strtoul(name, NULL, 0);
    }
    if (errno == ERANGE)
      return 0;

    switch (confel->dtype) {
      
    case DTYPE_INT:
      old = *(int*)((char*)conf + confel->offset);
      *(int*)((char*)conf + confel->offset) = tmp;
      break;
      
    case DTYPE_UINT:
      old = *(uint*)((char*)conf + confel->offset);
      *(uint*)((char*)conf + confel->offset) = tmp;
      break;
      
    case DTYPE_LONG:
      old = *(long*)((char*)conf + confel->offset);
      *(long*)((char*)conf + confel->offset) = tmp;
      break;
      
    case DTYPE_ULONG:
      old = *(ulong*)((char*)conf + confel->offset);
      *(ulong*)((char*)conf + confel->offset) = tmp;
      break;
      
    case DTYPE_SHORT:
      old = *(short*)((char*)conf + confel->offset);
      *(short*)((char*)conf + confel->offset) = tmp;
      break;
      
    case DTYPE_USHORT:
      old = *(ushort*)((char*)conf + confel->offset);
      *(ushort*)((char*)conf + confel->offset) = tmp;
      break;
      
    case DTYPE_CHAR:
      old = *(char*)((char*)conf + confel->offset);
      *(char*)((char*)conf + confel->offset) = tmp;
      break;
      
    case DTYPE_UCHAR:
      old = *(unsigned char*)((char*)conf + confel->offset);
      *(unsigned char*)((char*)conf + confel->offset) = tmp;
      break;
      
    case DTYPE_PTR:
      (void*)old = *(void**)((char*)conf + confel->offset);
      *(void**)((char*)conf + confel->offset) = (void*)tmp;
      break;
    }
    break;
  }
  if (old != tmp && confel->set_hook)
    (confel->set_hook)(name, confel, portp);
  return 1;
}

int nsg_port_get_name(char *name, 
		       nsg_port_config_t *confel, struct nsg_port_t *portp) {
  nsgdev_conf_t *conf = &portp->conf;
  uint tmp = 0;

  switch (confel->dtype) {
    
  case DTYPE_CONFIG:
    return 0;
    break;
    
  default:

    switch (confel->dtype) {
    case DTYPE_INT:
      tmp = *(int*)((char*)conf + confel->offset);
      break;
      
    case DTYPE_UINT:
      tmp = *(uint*)((char*)conf + confel->offset);
      break;
      
    case DTYPE_LONG:
      tmp = *(long*)((char*)conf + confel->offset);
      break;
      
    case DTYPE_ULONG:
      tmp = *(ulong*)((char*)conf + confel->offset);
      break;
      
    case DTYPE_SHORT:
      tmp = *(short*)((char*)conf + confel->offset);
      break;
      
    case DTYPE_USHORT:
      tmp = *(ushort*)((char*)conf + confel->offset);
      break;
      
    case DTYPE_CHAR:
      tmp = *(char*)((char*)conf + confel->offset);
      break;
      
    case DTYPE_UCHAR:
      tmp = *(unsigned char*)((char*)conf + confel->offset);
      break;
      
    case DTYPE_PTR:
      (void*)tmp = *(void**)((char*)conf + confel->offset);
      break;
    }
    if (confel->value) {
      return get_name_by_value(name, tmp, confel->value);
    }
    else {
      sprintf(name, "%u", tmp);
      return 1;
    }
    break;
  }
  return 0;
}


int set_param(struct vty *vty, int argc, char **argv, char *conf_name ) {
  struct nsg_port_t *portp = (struct nsg_port_t*) vty->index;
  int ret;

  ret = nsg_port_set_param(argc, argv,
			   nsg_port_get_config(conf_name, nsg_port_conf,
					       portp),
			   portp);

  if (!ret) {
    vty_out (vty, "Can't set parameter%s", VTY_NEWLINE);
    return CMD_WARNING;
  }
  else
    return CMD_SUCCESS;
}


DEFUN (nsg_phy_interface,
       nsg_phy_interface_cmd,
       PHY_INTERFACE_STR" ("V24_STR"|"V35_STR"|"SDSL_STR")",
       "Define physical interface\n"
       "V24 interface\n"
       "V35 interface\n"
       "SDSL interface"
       )
{
  return set_param(vty, argc, argv, PHY_INTERFACE_STR);
}

DEFUN (nsg_clock,
       nsg_clock_cmd,
       CLOCK_STR" ("EXTERNAL_STR"|"INTERNAL_STR"|"MASTER_STR"|"SLAVE_STR")",
       "Configure serial interface clock\n"
       "External mode\n"
       "Internal mode\n"
       "Master mode\n"
       "Slave mode\n"
       )
{
  return set_param(vty, argc, argv, CLOCK_STR);
}

ALIAS (nsg_clock,
       no_nsg_clock_cmd,
       "no "CLOCK_STR,
       NO_STR
       "Configure serial interface clock\n"
       );

DEFUN (nsg_baudrate,
       nsg_baudrate_cmd,
       BAUDRATE_STR" <300-2048000>",
       "Configure serial interface clock rate\n"
       "speed")
{
  return set_param(vty, argc, argv, BAUDRATE_STR);
}

DEFUN (nsg_port_encapsulation, 
       nsg_port_encapsulation_cmd,
       ENCAPSULATION_STR " (" CISCO_HDLC_STR "|" FRAME_RELAY_STR ")",
       "Set encapsulation type for an interface\n"
       	"Cisco-HDLC\n"
	"Frame Relay\n"
)
{
  return set_param(vty, argc, argv, ENCAPSULATION_STR);
}

DEFUN (nsg_if_cisco_hdlc, 
       nsg_if_cisco_hdlc_cmd,
       CISCO_HDLC_STR" ("KEEPALIVE_STR"|"TIMEOUT_STR") <1-100>",
       "set cisco HDLC parameters\n"
       "Keepalive period\n"
       "")
{
  return set_param(vty, argc, argv, CISCO_HDLC_STR);
}

ALIAS (nsg_if_cisco_hdlc, 
       no_nsg_if_cisco_hdlc_cmd,
       "no "CISCO_HDLC_STR" ("KEEPALIVE_STR"|"TIMEOUT_STR")",
       NO_STR
       "set cisco HDLC parameters\n"
       "Keepalive period (default 10 seconds)\n"
       "");


DEFUN (nsg_if_frame_relay_tn, 
       nsg_if_frame_relay_tn_cmd,
       FRAME_RELAY_STR" ("T391_STR"|"T392_STR"|"N391_STR"|"N392_STR"|"N393_STR
       ") <0-255>",
       "Set Frame Relay parameters\n"
       "Link integrity verification polling timer\n"
       "Polling verification timer\n"
       "Full status polling counter\n"
       "LMI error threshold\n"
       "LMI monitored event count"
       )
{
  return set_param(vty, argc, argv, FRAME_RELAY_STR);
}

ALIAS (nsg_if_frame_relay_tn, 
       no_nsg_if_frame_relay_tn_cmd,
       "no "FRAME_RELAY_STR" ("T391_STR"|"T392_STR"|"N391_STR"|"N392_STR"|"N393_STR
       ")",
	NO_STR
       "Set Frame Relay parameters\n"
       "Link integrity verification polling timer\n" \
       "Polling verification timer\n" \
       "Full status polling counter\n" \
       "LMI error threshold\n" \
       "LMI monitored event count"
       )

DEFUN (nsg_if_frame_relay_lmi, 
       nsg_if_frame_relay_lmi_cmd,
       FRAME_RELAY_STR" "LMI_STR" ("NONE_STR"|"ANSI_STR"|"CCITT_STR")",
       "set Frame Relay parameters\n"
       "Use NONE-ANSI-CCITT type LMI")
{
  return set_param(vty, argc, argv, FRAME_RELAY_STR" "LMI_STR);
}

ALIAS (nsg_if_frame_relay_lmi, 
       no_nsg_if_frame_relay_lmi_cmd,
       "no "FRAME_RELAY_STR" "LMI_STR,
       NO_STR
       "set Frame Relay parameters\n"
       "Use NONE-ANSI-CCITT type LMI")

DEFUN (nsg_if_frame_relay_intf_type, 
       nsg_if_frame_relay_intf_type_cmd,
       FRAME_RELAY_STR" "INTF_TYPE_STR" ("DTE_STR"|"DCE_STR")",
       "set Frame Relay parameters\n"
       "Use DTE-DCE interface type LMI")
{
  return set_param(vty, argc, argv, FRAME_RELAY_STR" "INTF_TYPE_STR);
}

int nsg_port_ctl(struct nsg_port_t *portp,
		 struct nsg_channel_t *channelp,
		 int ioctl_code) {
  int ret, fd;
  char filename[sizeof(NSG_ROUTER_PORT_DIR) + NSG_DRVNAME_SZ + 2];
  void *conf;

  sprintf(filename, "%s/%s", NSG_ROUTER_PORT_DIR, portp->name);
  fd = open(filename, O_RDONLY);

  conf = (ioctl_code == NSG_IFDEL || ioctl_code == NSG_IFNEW) ? &channelp->conf :
    (void *)&portp->conf;

  if ((fd < 0) || (ioctl(fd, ioctl_code, conf) < 0)) {
    ret = CMD_WARNING;
  }
  else {
    if (ioctl_code == NSG_SETUP) {
      /* setup all interfaces */
      listnode node;
      struct nsg_channel_t *channelp;
      for (node = listhead (portp->channel); node; nextnode (node)) {
	channelp = getdata (node);
	if (CHECK_FLAG(channelp->flags, NSG_FLG_NOSHUTDOWN)) {
	  if (ioctl(fd, NSG_IFNEW, &channelp->conf) < 0) {
	  }
	}
      }
    }

    ret = CMD_SUCCESS;
  }
  if(fd >= 0) 
    close (fd);
  return ret;
}

DEFUN (nsg_dlci_shutdown,
       nsg_dlci_shutdown_cmd,
       "shutdown",
       "Shutdown DLCI")
{
  int ret;
  struct nsg_channel_t *channelp = (struct nsg_channel_t*) vty->index;
  struct nsg_port_t *portp = channelp->portp;

  UNSET_FLAG(channelp->flags, NSG_FLG_NOSHUTDOWN);

  if ((ret = nsg_port_ctl(portp, channelp, NSG_IFDEL)) != CMD_SUCCESS) {
    vty_out (vty, "Can't shutdown DLCI%s", VTY_NEWLINE);
  }
  return ret;
}

DEFUN (no_nsg_dlci_shutdown,
       no_nsg_dlci_shutdown_cmd,
       "no shutdown",
       "Setup DLCI")
{
  int ret;
  struct nsg_channel_t *channelp = (struct nsg_channel_t*) vty->index;
  struct nsg_port_t *portp = channelp->portp;

  SET_FLAG(channelp->flags, NSG_FLG_NOSHUTDOWN);

  if ((ret = nsg_port_ctl(portp, channelp, NSG_IFNEW)) != CMD_SUCCESS) {
    vty_out (vty, "Can't setup DLCI%s", VTY_NEWLINE);
  }
  return ret;
}

DEFUN (nsg_port_shutdown,
       nsg_port_shutdown_cmd,
       "shutdown",
       "Shutdown port")
{
  int ret;
  struct nsg_port_t *portp = (struct nsg_port_t*) vty->index;

  UNSET_FLAG(portp->flags, NSG_FLG_NOSHUTDOWN);

  if ((ret = nsg_port_ctl(portp, NULL, NSG_DOWN)) != CMD_SUCCESS) {
    vty_out (vty, "Can't shutdown port%s", VTY_NEWLINE);
  }
  return ret;
}

DEFUN (no_nsg_port_shutdown,
       no_nsg_port_shutdown_cmd,
       "no shutdown",
       NO_STR
       "Setup port")
{
  int ret;
  struct nsg_port_t *portp = (struct nsg_port_t*) vty->index;

  SET_FLAG(portp->flags, NSG_FLG_NOSHUTDOWN);

  if ((ret = nsg_port_ctl(portp, NULL, NSG_SETUP)) != CMD_SUCCESS) {
    vty_out (vty, "Can't setup port%s", VTY_NEWLINE);
  }
  return ret;
}

/* Create new nsg_channel structure. */
struct nsg_channel_t *
nsg_channel_new ()
{
  struct nsg_channel_t *channelp;

  channelp = XMALLOC (MTYPE_NSG_CHANNEL, sizeof (struct nsg_channel_t));
  memset (channelp, 0, sizeof (struct nsg_channel_t));
  return channelp;
}

struct nsg_channel_t *
nsg_channel_create (struct nsg_port_t *portp)
{
  struct nsg_channel_t *channelp;

  channelp = nsg_channel_new ();
  
  listnode_add (portp->channel, channelp);

  channelp->conf.magic = NSG_MAGIC;

  return channelp;
}

/* Delete and free interface structure. */
void
nsg_channel_free (struct nsg_channel_t *channelp)
{
  XFREE (MTYPE_NSG_CHANNEL, channelp);
}

void
nsg_channel_delete (struct nsg_port_t *portp, struct nsg_channel_t *channelp)
{
  listnode_delete (portp->channel, channelp);
  nsg_channel_free(channelp);
}




struct nsg_port_master
{
  int (*nsg_port_new_hook) (struct nsg_port_t *);
  int (*nsg_port_delete_hook) (struct nsg_port_t *);
} nsg_port_master;



/* Create new nsg_port structure. */
struct nsg_port_t *
nsg_port_new ()
{
  struct nsg_port_t *portp;

  portp = XMALLOC (MTYPE_NSG_PORT, sizeof (struct nsg_port_t));
  memset (portp, 0, sizeof (struct nsg_port_t));
  return portp;
}

struct nsg_port_t *
nsg_port_create ()
{
  struct nsg_port_t *portp;

  portp = nsg_port_new ();
  
  listnode_add (nsg_port_list, portp);

  portp->channel = list_new ();
  portp->channel->del = (void (*) (void *)) nsg_channel_free;


  if (nsg_port_master.nsg_port_new_hook)
    (*nsg_port_master.nsg_port_new_hook) (portp);

  return portp;
}

/* Delete and free interface structure. */
void
nsg_port_delete (struct nsg_port_t *portp)
{
  listnode_delete (nsg_port_list, portp);

  if (nsg_port_master.nsg_port_delete_hook)
    (*nsg_port_master.nsg_port_delete_hook) (portp);

  /* Free channel list */
  list_delete (portp->channel);

  XFREE (MTYPE_NSG_PORT, portp);
}



/* Add hook to interface master. */
void
nsg_port_add_hook (int type, int (*func)(struct nsg_port_t *portp))
{
  switch (type) {
  case NSG_PORT_NEW_HOOK:
    nsg_port_master.nsg_port_new_hook = func;
    break;
  case NSG_PORT_DELETE_HOOK:
    nsg_port_master.nsg_port_delete_hook = func;
    break;
  default:
    break;
  }
}

/* nsg_port existance check by nsg_port name. */
struct nsg_port_t *
nsg_port_lookup_by_name (char *name)
{
  listnode node;
  struct nsg_port_t *portp;

  for (node = listhead (nsg_port_list); node; nextnode (node))
    {
      portp = getdata (node);
      if (strncmp (name, portp->name, sizeof portp->name) == 0)
	return portp;
    }
  return NULL;
}

int is_nsg_port(char *name)
{
  int fd;
  char filename[sizeof(NSG_ROUTER_PORT_DIR) + NSG_DRVNAME_SZ + 2];
  sprintf(filename, "%s/%s", NSG_ROUTER_PORT_DIR, name);

  if ((fd = open(filename, O_RDONLY)) < 0) {
    return 0;
  } 
  else {
    close(fd);
    return 1;
  }
}

void nsg_port_set_default(nsg_port_config_t *conftable,
			  struct nsg_port_t *portp) {
  nsg_port_config_t *p;
  
  for (p = conftable; p->keyword; p++) {
    if (!nsg_port_test_cond(p->cond, nsg_port_conf, portp))
      continue;

    if (p->dtype == DTYPE_CONFIG) {
      nsg_port_set_default(p->value, portp);
    }
    else {
      nsg_port_set_param(0, NULL, p, portp);
    }
  }
}

void set_encap_param_hook(char *name, nsg_port_config_t *confel, struct nsg_port_t *portp) {
  nsg_port_config_t *encap_config;

  if (confel->value) {
    name = get_full_name(name, strlen(name), confel->value);
    if (name) {
      encap_config = nsg_port_get_config(name, nsg_port_conf, portp);
      if (encap_config->value)
	nsg_port_set_default(encap_config->value, portp);

      list_delete_all_node(portp->channel);

      if ((strcmp(name, CISCO_HDLC_STR)) == 0) {
	struct nsg_channel_t *channelp = nsg_channel_create(portp);
	channelp->conf.encap = portp->conf.encap;
	SET_FLAG(channelp->flags, NSG_FLG_NOSHUTDOWN);
      }
    }
  }
}


int nsg_port_new_hook(struct nsg_port_t *portp)
{
  portp->conf.magic = NSG_MAGIC;
  nsg_port_set_default(nsg_port_conf, portp);
  return 0;
}

int nsg_port_delete_hook(struct nsg_port_t *portp)
{
  return 0;
}

DEFUN_NOSH (nsg_port,
	    nsg_port_cmd,
	    "port PORTNAME",
	    "Select a port to configure\n"
	    "Port's name\n")
{
  struct nsg_port_t *portp;

  portp = nsg_port_lookup_by_name (argv[0]);

  if (portp == NULL)
    {
      if (!is_nsg_port(argv[0])) {
	vty_out (vty, "No such port%s", VTY_NEWLINE);
	return CMD_WARNING;
      }
      portp = nsg_port_create();
      strncpy (portp->name, argv[0], NSG_IFNAME_SZ);
    }
  vty->index = portp;
  vty->node = NSG_PORT_NODE;

  return CMD_SUCCESS;
}

DEFUN (no_nsg_port,
       no_nsg_port_cmd,
       "no port PORTNAME",
       NO_STR
       "Delete port configuration\n"
       "Port's name\n")
{
  struct nsg_port_t *portp;

  portp = nsg_port_lookup_by_name (argv[0]);

  if (portp && is_nsg_port(argv[0]))
    {
      nsg_port_ctl(portp, NULL, NSG_DOWN);
      nsg_port_delete(portp);
      return CMD_SUCCESS;
    }
  else {
    vty_out (vty, "No such port%s", VTY_NEWLINE);
    return CMD_WARNING;
  }
}

void nsg_port_write_all_conf(struct vty *vty, char *prompt,
			     nsg_port_config_t *conftable,
			     struct nsg_port_t *portp) {
  nsg_port_config_t *p;

  for (p = conftable; p->keyword; p++) {
    if (!nsg_port_test_cond(p->cond, nsg_port_conf, portp))
      continue;

    if (p->dtype == DTYPE_CONFIG) {
      char newprompt[strlen(prompt) + strlen(p->keyword) + 2];
      strcpy(newprompt, prompt);
      strcat(newprompt, p->keyword);
      strcat(newprompt, " ");

      nsg_port_write_all_conf(vty, newprompt, p->value, portp);
    }
    else {
      if (!nsg_port_test_param(0, p, portp)) {
	char name[128];
	if (!nsg_port_get_name(name, p, portp)) {
	  strcpy(name, "<UNKNOWN>");
	}
	vty_out(vty, "%s%s %s%s",
		prompt, p->keyword, name, VTY_NEWLINE);
      }
    }
  }
}

static void nsg_fr_dlci_write (struct vty *vty,
			      struct nsg_port_t *portp,
			      struct nsg_channel_t *channelp) {
    vty_out(vty, " "FRAME_RELAY_STR" "DLCI_STR" %d%s", channelp->conf.param.fr.dlci,
	     VTY_NEWLINE);

    if (CHECK_FLAG(channelp->flags, NSG_FLG_NOSHUTDOWN)) {
      vty_out (vty, "  no shutdown%s",
	       VTY_NEWLINE);
    }
}

int nsg_fr_dlci_config_write (struct vty *vty, struct nsg_port_t *portp) {
  listnode node;
  
  for (node = listhead (portp->channel); node; nextnode (node)) {
    struct nsg_channel_t *channelp = getdata (node);
    nsg_fr_dlci_write(vty, portp, channelp);  
    
    vty_out(vty, "  exit%s", VTY_NEWLINE);
  }  
  return 0;
}
 
static void nsg_port_write(struct vty *vty, struct nsg_port_t *portp ) {
    vty_out (vty, "port %s%s", portp->name,
	     VTY_NEWLINE);
    
    nsg_port_write_all_conf(vty, " ", nsg_port_conf, portp);
    
    if CHECK_FLAG(portp->flags, NSG_FLG_NOSHUTDOWN) {
      vty_out (vty, " no shutdown%s",
	       VTY_NEWLINE);
    }
    
    if (nsg_port_test_cond(ENCAPSULATION_STR" "FRAME_RELAY_STR, nsg_port_conf, portp)) {
      nsg_fr_dlci_config_write(vty, portp);
    }
    
    vty_out (vty, "!%s", VTY_NEWLINE);
}

int nsg_port_config_write (struct vty *vty) {
  listnode node;
  
  for (node = listhead (nsg_port_list); node; nextnode (node)) {
    struct nsg_port_t *portp = getdata (node);
    nsg_port_write(vty, portp);
  }
  return 0;
}

DEFUN (nsg_port_write_term,
       nsg_port_write_term_cmd,
       "write",
       "Write port configuration")
{
  struct nsg_port_t *portp = (struct nsg_port_t*) vty->index;

  nsg_port_write(vty, portp);
  return CMD_SUCCESS;
}

DEFUN (nsg_dlci_write,
       nsg_dlci_write_cmd,
       "write",
       "Write dlci configuration")
{
  struct nsg_channel_t *channelp = (struct nsg_channel_t*) vty->index;

  nsg_fr_dlci_write(vty, channelp->portp, channelp);
  return CMD_SUCCESS;
}

struct cmd_node nsg_port_node =
{
  NSG_PORT_NODE,
  "%s(config-port)# ",
  1
};

/* Initialize nsg_port list. */
void
nsg_port_init ()
{
  nsg_port_list = list_new ();

  if (nsg_port_list)
    return;

  memset (&nsg_port_master, 0, sizeof nsg_port_master);
}

struct nsg_channel_t *
nsg_channel_lookup_by_dlci(struct nsg_port_t *portp, ushort dlci)
{
  listnode node;
  struct nsg_channel_t *channelp;

  for (node = listhead (portp->channel); node; nextnode (node))
    {
      channelp = getdata (node);
      if (channelp->conf.encap == 
	  get_value_by_name(FRAME_RELAY_STR,
			    strlen(FRAME_RELAY_STR),
			    nsg_port_encap_value)  &&
	  channelp->conf.param.fr.dlci == dlci)
	return channelp;
    }
  return NULL;
}

DEFUN_NOSH (nsg_fr_dlci,
	    nsg_fr_dlci_cmd,
	    FRAME_RELAY_STR" "DLCI_STR" <16-1022>",
	    "Select a frame dlci to configure\n"
	    "DLCI\n")
{
  struct nsg_port_t *portp = vty->index;
  struct nsg_channel_t *channelp;
  ushort dlci;

  if (!nsg_port_test_cond(ENCAPSULATION_STR" "FRAME_RELAY_STR, nsg_port_conf, portp)) {
    vty_out (vty, ENCAPSULATION_STR" not "FRAME_RELAY_STR"%s", VTY_NEWLINE);
    return CMD_WARNING;
  }

  dlci = strtoul(argv[0], NULL, 0);
  channelp = nsg_channel_lookup_by_dlci(portp, dlci);

  if (channelp == NULL) {
    channelp = nsg_channel_create(portp);
    channelp->conf.encap = portp->conf.encap;
    channelp->conf.param.fr.dlci = dlci;
    channelp->portp = portp;  
    snprintf(channelp->conf.name, sizeof(channelp->conf.name),
	     "%s/%d", portp->name, dlci);

  }
  
  vty->index = channelp;
  vty->node = NSG_FR_DLCI_NODE;
  return CMD_SUCCESS;
}

DEFUN(no_nsg_fr_dlci,
      no_nsg_fr_dlci_cmd,
      "no "FRAME_RELAY_STR" "DLCI_STR" <16-1022>",
      NO_STR
      "Delete a frame relay dlci configuration\n"
      "DLCI\n")
{
  struct nsg_port_t *portp = vty->index;
  struct nsg_channel_t *channelp;
  ushort dlci;

  if (!nsg_port_test_cond(ENCAPSULATION_STR" "FRAME_RELAY_STR, nsg_port_conf, portp)) {
    vty_out (vty, ENCAPSULATION_STR" not "FRAME_RELAY_STR"%s", VTY_NEWLINE);
    return CMD_WARNING;
  }

  dlci = strtoul(argv[0], NULL, 0);
  channelp = nsg_channel_lookup_by_dlci(portp, dlci);

  if (channelp) {
    nsg_channel_delete(portp, channelp);
  }
  
  return CMD_SUCCESS;
}

struct cmd_node nsg_fr_dlci_node =
{
  NSG_FR_DLCI_NODE,
  "%s(config-fr-dlci)# ",
  1
};

void nsg_install_default (enum node_type node)
{
  install_element (node, &config_exit_cmd);
  install_element (node, &config_quit_cmd);
  install_element (node, &config_end_cmd);
  install_element (node, &config_help_cmd);
  install_element (node, &config_list_cmd);
}

void
vtysh_init_nsg_cmd ()
{

  /* Initialize nsg_port and new hook. */
  nsg_port_init();
  nsg_port_add_hook (NSG_PORT_NEW_HOOK, nsg_port_new_hook);
  nsg_port_add_hook (NSG_PORT_DELETE_HOOK, nsg_port_delete_hook);

  install_element (CONFIG_NODE, &nsg_port_cmd);
  install_element (CONFIG_NODE, &no_nsg_port_cmd);
  install_node (&nsg_port_node, nsg_port_config_write);
  nsg_install_default (NSG_PORT_NODE);

  install_element (NSG_PORT_NODE, &nsg_fr_dlci_cmd);
  install_element (NSG_PORT_NODE, &no_nsg_fr_dlci_cmd);
  install_node (&nsg_fr_dlci_node, NULL);
  nsg_install_default (NSG_FR_DLCI_NODE);

  install_element (NSG_PORT_NODE, &nsg_phy_interface_cmd);
  install_element (NSG_PORT_NODE, &nsg_clock_cmd);
  install_element (NSG_PORT_NODE, &no_nsg_clock_cmd);
  install_element (NSG_PORT_NODE, &nsg_baudrate_cmd);
  install_element (NSG_PORT_NODE, &nsg_port_encapsulation_cmd);
  install_element (NSG_PORT_NODE, &nsg_if_cisco_hdlc_cmd);
  install_element (NSG_PORT_NODE, &no_nsg_if_cisco_hdlc_cmd);
  install_element (NSG_PORT_NODE, &nsg_if_frame_relay_tn_cmd);
  install_element (NSG_PORT_NODE, &no_nsg_if_frame_relay_tn_cmd);
  install_element (NSG_PORT_NODE, &nsg_if_frame_relay_lmi_cmd);
  install_element (NSG_PORT_NODE, &nsg_if_frame_relay_intf_type_cmd);
  install_element (NSG_PORT_NODE, &no_nsg_if_frame_relay_lmi_cmd);
  install_element (NSG_PORT_NODE, &nsg_port_shutdown_cmd);
  install_element (NSG_PORT_NODE, &no_nsg_port_shutdown_cmd);
  install_element (NSG_PORT_NODE, &nsg_port_write_term_cmd);
  install_element (NSG_FR_DLCI_NODE, &nsg_dlci_shutdown_cmd);
  install_element (NSG_FR_DLCI_NODE, &no_nsg_dlci_shutdown_cmd);
  install_element (NSG_FR_DLCI_NODE, &nsg_dlci_write_cmd);
}
